@@ -1,6 +1,6 @@
 # -*- coding: UTF-8 -*-
 
-# Copyright (C) 2005, 2006, 2007 Canonical Ltd.
+# Copyright (C) 2005, 2006, 2007, 2008, 2009 Canonical Ltd.
 # Copyright (C) 2007 Mario Limonciello <superm1@ubuntu.com>
 # Written by Colin Watson <cjwatson@ubuntu.com>.
 #
@@ -29,35 +29,18 @@ import gobject
 import gtk
 import gtk.glade
 
-from debconf import DebconfCommunicator
-
-from oem_config import filteredcommand, i18n, zoommap
+from oem_config import filteredcommand, i18n, timezone_map
 from oem_config.components import console_setup, language, timezone, user, \
                                   language_apply, timezone_apply, \
                                   console_setup_apply
 import oem_config.emap
 import oem_config.tz
+from oem_config.frontend.base import BaseFrontend
 
 PATH = '/usr/share/oem-config'
 GLADEDIR = '/usr/share/oem-config/glade'
 
-BREADCRUMB_STEPS = {
-    "step_language": 1,
-    "step_timezone": 2,
-    "step_keyboard": 3,
-    "step_user": 4,
-}
-BREADCRUMB_MAX_STEP = 4
-
-# Define what pages of the UI we want to load.
-SUBPAGES = [
-    "step_language",
-    "step_timezone",
-    "step_keyboard",
-    "step_user",
-]
-
-class Frontend:
+class Frontend(BaseFrontend):
     def __init__(self):
         def add_subpage(self, steps, name):
             """Inserts a subpage into the notebook.  This assumes the file
@@ -83,16 +66,15 @@ class Frontend:
                 if isinstance(widget, gtk.Label):
                     widget.set_property('can-focus', False)
 
+        BaseFrontend.__init__(self)
+
         self.previous_excepthook = sys.excepthook
         sys.excepthook = self.excepthook
 
         self.all_widgets = set()
         self.language_questions = ('oem_config', 'language_heading_label',
-                                   'language_text_label', 'step_label',
-                                   'back', 'next')
+                                   'step_label', 'back', 'next')
         self.current_page = None
-        self.locale = None
-        self.current_layout = None
         self.username_edited = False
         self.allowed_change_step = True
         self.allowed_go_forward = True
@@ -100,8 +82,7 @@ class Frontend:
         self.apply_changes = False
 
         # Set default language.
-        dbfilter = language.Language(self, DebconfCommunicator('oem-config',
-                                                               cloexec=True))
+        dbfilter = language.Language(self, self.debconf_communicator())
         dbfilter.cleanup()
         dbfilter.db.shutdown()
 
@@ -113,46 +94,79 @@ class Frontend:
         add_widgets(self, self.glade)
 
         steps = self.glade.get_widget("steps")
-        for page in SUBPAGES:
+        for page in self.pages:
             add_subpage(self, steps, page)
 
+        def win_size_req(win, req):
+            s = win.get_screen()
+            m = s.get_monitor_geometry(0)
+            w = -1
+            h = -1
+
+            # What's the size of the WM border?
+            total_frame = win.window.get_frame_extents()
+            (cur_x, cur_y, cur_w, cur_h, depth) = win.window.get_geometry()
+            wm_w = total_frame.width - cur_w
+            wm_h = total_frame.height - cur_h
+
+            if req.width > m.width - wm_w:
+                w = m.width - wm_w
+            if req.height > m.height - wm_h:
+                h = m.height - wm_h
+
+            win.set_size_request(w, h)
+            win.resize(w, h)
+
+        if 'OEM_CONFIG_ONLY' in os.environ:
+            self.oem_config.fullscreen()
+        else:
+            self.oem_config.connect('size-request', win_size_req)
+
         self.translate_widgets()
 
         if 'UBIQUITY_OLD_TZMAP' in os.environ:
             self.tzmap = TimezoneMap(self)
             self.tzmap.tzmap.show()
         else:
-            pixmap = '/usr/share/oem-config/pixmaps/earth.jpg'
-            full_zoom = True
-            font_selected = "white"
-            font_unselected = "orange"
-            args = (self, pixmap, full_zoom, font_selected, font_unselected)
-            self.tzmap = zoommap.ZoomMapWidget(*args)
+            self.tzdb = oem_config.tz.Database()
+            self.tzmap = timezone_map.TimezoneMap(self.tzdb,
+                '/usr/share/oem-config/pixmaps/timezone')
+            self.tzmap.connect('city-selected', self.select_city)
+            self.timezone_map_window.add(self.tzmap)
+            self.setup_timezone_page()
             self.tzmap.show()
 
         if 'OEM_CONFIG_DEBUG' in os.environ:
             self.password_debug_warning_label.show()
 
-    def post_mortem(self, exctype, excvalue, exctb):
-        """Drop into the debugger if possible."""
+    def setup_timezone_page(self):
+        renderer = gtk.CellRendererText()
+        self.timezone_zone_combo.pack_start(renderer, True)
+        self.timezone_zone_combo.add_attribute(renderer, 'text', 0)
+        list_store = gtk.ListStore(gobject.TYPE_STRING)
+        self.timezone_zone_combo.set_model(list_store)
+        self.timezone_zone_combo.connect('changed',
+            self.zone_combo_selection_changed)
+        self.timezone_city_combo.connect('changed',
+            self.city_combo_selection_changed)
 
-        # Did the user request this?
-        if 'OEM_CONFIG_DEBUG_PDB' not in os.environ:
-            return
-        # We must not be in interactive mode; if we are, there's no point.
-        if hasattr(sys, 'ps1'):
-            return
-        # stdin and stdout must point to a terminal. (stderr is redirected
-        # in debug mode!)
-        if not sys.stdin.isatty() or not sys.stdout.isatty():
-            return
-        # SyntaxErrors can't meaningfully be debugged.
-        if issubclass(exctype, SyntaxError):
-            return
+        renderer = gtk.CellRendererText()
+        self.timezone_city_combo.pack_start(renderer, True)
+        self.timezone_city_combo.add_attribute(renderer, 'text', 0)
+        city_store = gtk.ListStore(gobject.TYPE_STRING)
+        self.timezone_city_combo.set_model(city_store)
 
-        import pdb
-        pdb.post_mortem(exctb)
-        sys.exit(1)
+        self.regions = {}
+        for location in self.tzdb.locations:
+            region, city = location.zone.replace('_', ' ').split('/', 1)
+            if region in self.regions:
+                self.regions[region].append(city)
+            else:
+                self.regions[region] = [city]
+
+        r = self.regions.keys()
+        for region in r:
+            list_store.append([region])
 
     def excepthook(self, exctype, excvalue, exctb):
         """Crash handler."""
@@ -238,12 +252,10 @@ class Frontend:
 
         self.oem_config.hide()
 
-    def set_window_hints(self, widget):
-        if 'OEM_CONFIG_ONLY' in os.environ:
-            # Disable minimise button.
-            widget.window.set_functions(
-                gtk.gdk.FUNC_RESIZE | gtk.gdk.FUNC_MOVE |
-                gtk.gdk.FUNC_MAXIMIZE)
+        if self.apply_changes:
+            return 0
+        else:
+            return 10
 
     # Internationalisation.
 
@@ -272,14 +284,15 @@ class Frontend:
 
         if isinstance(widget, gtk.Label):
             if name == 'step_label':
-                global BREADCRUMB_STEPS, BREADCRUMB_MAX_STEP
                 curstep = '?'
                 if self.current_page is not None:
-                    current_name = self.step_name(self.current_page)
-                    if current_name in BREADCRUMB_STEPS:
-                        curstep = str(BREADCRUMB_STEPS[current_name])
+                    # TODO cjwatson 2009-01-20: This is a compromise. It
+                    # doesn't work with multi-page steps (e.g. partitioning
+                    # in ubiquity), but it simplifies page configuration
+                    # quite a bit.
+                    curstep = str(self.current_page + 1)
                 text = text.replace('${INDEX}', curstep)
-                text = text.replace('${TOTAL}', str(BREADCRUMB_MAX_STEP))
+                text = text.replace('${TOTAL}', str(self.steps.get_n_pages()))
             widget.set_text(text)
 
             # Ideally, these attributes would be in the glade file somehow ...
@@ -314,12 +327,6 @@ class Frontend:
         elif isinstance(widget, gtk.Window):
             widget.set_title(text)
 
-    def get_string(self, name, lang=None):
-        """Get the string name in the given lang or a default."""
-        if lang is None:
-            lang = self.locale
-        return i18n.get_string(name, lang)
-
     # I/O helpers.
 
     def watch_debconf_fd (self, from_debconf, process_input):
@@ -350,6 +357,40 @@ class Frontend:
         dialog.run()
         dialog.hide()
 
+    def question_dialog (self, title, msg, options, use_templates=True):
+        self.allow_change_step(True)
+        if not msg:
+            msg = title
+        buttons = []
+        for option in options:
+            if use_templates:
+                text = self.get_string(option)
+            else:
+                text = option
+            if text is None:
+                text = option
+            # Work around PyGTK bug; each button text must actually be a
+            # subtype of str, which unicode isn't.
+            text = str(text)
+            buttons.extend((text, len(buttons) / 2 + 1))
+        dialog = gtk.Dialog(title, self.oem_config,
+                            gtk.DIALOG_MODAL, tuple(buttons))
+        vbox = gtk.VBox()
+        vbox.set_border_width(5)
+        label = gtk.Label(msg)
+        label.set_line_wrap(True)
+        label.set_selectable(True)
+        vbox.pack_start(label)
+        vbox.show_all()
+        dialog.vbox.pack_start(vbox)
+        response = dialog.run()
+        dialog.hide()
+        if response < 0:
+            # something other than a button press, probably destroyed
+            return None
+        else:
+            return options[response - 1]
+
     # Run the UI's main loop until it returns control to us.
     def run_main_loop(self):
         if not self.apply_changes:
@@ -366,7 +407,6 @@ class Frontend:
         return self.steps.get_nth_page(step_index).get_name()
 
     def set_current_page(self, current):
-        global BREADCRUMB_STEPS, BREADCRUMB_MAX_STEP
         self.current_page = current
         self.translate_widget(self.step_label, self.locale)
         if current == 0:
@@ -407,8 +447,7 @@ class Frontend:
         self.allowed_go_forward = allowed
 
     def debconffilter_done(self, dbfilter):
-        if dbfilter == self.dbfilter:
-            self.dbfilter = None
+        if BaseFrontend.debconffilter_done(self, dbfilter):
             gtk.main_quit()
 
     def on_back_clicked(self, widget):
@@ -450,54 +489,96 @@ class Frontend:
 
     # Callbacks provided to components.
 
-    def selected_language(self, selection):
-        (model, iterator) = selection.get_selected()
+    def zone_combo_selection_changed(self, widget):
+        i = self.timezone_zone_combo.get_active()
+        m = self.timezone_zone_combo.get_model()
+        region = m[i][0]
+
+        m = self.timezone_city_combo.get_model()
+        m.clear()
+        for city in self.regions[region]:
+            m.append([city])
+
+    def city_combo_selection_changed(self, widget):
+        i = self.timezone_zone_combo.get_active()
+        m = self.timezone_zone_combo.get_model()
+        region = m[i][0]
+
+        i = self.timezone_city_combo.get_active()
+        if i < 0:
+            # There's no selection yet.
+            return
+        m = self.timezone_city_combo.get_model()
+        city = m[i][0].replace(' ', '_')
+        city = region + '/' + city
+        self.tzmap.select_city(city)
+
+    def select_city(self, widget, city):
+        region, city = city.replace('_', ' ').split('/', 1)
+        m = self.timezone_zone_combo.get_model()
+        iterator = m.get_iter_first()
+        while iterator:
+            if m[iterator][0] == region:
+                self.timezone_zone_combo.set_active_iter(iterator)
+                break
+            iterator = m.iter_next(iterator)
+
+        m = self.timezone_city_combo.get_model()
+        iterator = m.get_iter_first()
+        while iterator:
+            if m[iterator][0] == city:
+                self.timezone_city_combo.set_active_iter(iterator)
+                break
+            iterator = m.iter_next(iterator)
+
+    def selected_language (self):
+        model = self.language_iconview.get_model()
+        items = self.language_iconview.get_selected_items()
+        if not items:
+            return ''
+        iterator = model.get_iter(items[0])
         if iterator is not None:
             value = unicode(model.get_value(iterator, 0))
             return self.language_choice_map[value][1]
-        else:
-            return ''
 
     def set_language_choices(self, choices, choice_map):
-        self.language_choice_map = dict(choice_map)
-        if len(self.language_treeview.get_columns()) < 1:
-            column = gtk.TreeViewColumn(None, gtk.CellRendererText(), text=0)
-            column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
-            self.language_treeview.append_column(column)
-            selection = self.language_treeview.get_selection()
-            selection.connect('changed',
-                              self.on_language_treeview_selection_changed)
+        BaseFrontend.set_language_choices(self, choices, choice_map)
+        self.language_iconview.connect('selection-changed',
+                          self.on_language_iconview_selection_changed)
         list_store = gtk.ListStore(gobject.TYPE_STRING)
-        self.language_treeview.set_model(list_store)
+        self.language_iconview.set_model(list_store)
+        self.language_iconview.set_text_column(0)
         for choice in choices:
             list_store.append([choice])
 
     def set_language(self, language):
-        model = self.language_treeview.get_model()
+        model = self.language_iconview.get_model()
         iterator = model.iter_children(None)
         while iterator is not None:
             if unicode(model.get_value(iterator, 0)) == language:
                 path = model.get_path(iterator)
-                self.language_treeview.get_selection().select_path(path)
-                self.language_treeview.scroll_to_cell(
-                    path, use_align=True, row_align=0.5)
+                self.language_iconview.select_path(path)
+                self.language_iconview.scroll_to_path(path, True, 0.5, 0.5)
                 break
             iterator = model.iter_next(iterator)
 
     def get_language(self):
-        selection = self.language_treeview.get_selection()
-        (model, iterator) = selection.get_selected()
+        model = self.language_iconview.get_model()
+        items = self.language_iconview.get_selected_items()
+        if not items:
+            return 'C'
+        iterator = model.get_iter(items[0])
         if iterator is None:
             return 'C'
         else:
             value = unicode(model.get_value(iterator, 0))
             return self.language_choice_map[value][1]
 
-    def on_language_treeview_row_activated(self, treeview, path, view_column):
+    def on_language_iconview_item_activated (self, iconview, path):
         self.next.activate()
 
-    def on_language_treeview_selection_changed(self, selection):
-        lang = self.selected_language(selection)
+    def on_language_iconview_selection_changed (self, iconview):
+        lang = self.selected_language()
         if lang:
             # strip encoding; we use UTF-8 internally no matter what
             lang = lang.split('.')[0].lower()
@@ -505,10 +586,18 @@ class Frontend:
                 self.translate_widget(getattr(self, widget), lang)
 
     def set_timezone (self, timezone):
-        self.tzmap.set_tz_from_name(timezone)
+        self.select_city(None, timezone)
 
     def get_timezone (self):
-        return self.tzmap.get_selected_tz_name()
+        i = self.timezone_zone_combo.get_active()
+        m = self.timezone_zone_combo.get_model()
+        region = m[i][0]
+
+        i = self.timezone_city_combo.get_active()
+        m = self.timezone_city_combo.get_model()
+        city = m[i][0].replace(' ', '_')
+        city = region + '/' + city
+        return city
 
     def set_keyboard_choices(self, choices):
         layouts = gtk.ListStore(gobject.TYPE_STRING)
@@ -528,7 +617,7 @@ class Frontend:
             self.set_keyboard(self.current_layout)
 
     def set_keyboard (self, layout):
-        self.current_layout = layout
+        BaseFrontend.set_keyboard(self, layout)
         model = self.keyboardlayoutview.get_model()
         if model is None:
             return
