@@ -1,6 +1,7 @@
-# -*- coding: UTF-8 -*-
+# -*- coding: utf-8 -*-
 
 # Copyright (C) 2006, 2007 Anirudh Ramesh
+# Copyright (C) 2006, 2007, 2008, 2009 Canonical Ltd.
 # Written by Anirudh Ramesh <abattoir@abattoir.in>.
 #
 # This program is free software; you can redistribute it and/or modify
@@ -31,25 +32,18 @@ from PyKDE4.kdeui import KApplication, KMainWindow, KIcon
 from PyQt4.QtCore import *
 from PyQt4.QtGui import *
 from PyQt4 import uic
-from debconf import DebconfCommunicator
 
 from oem_config import filteredcommand, i18n
 from oem_config.components import console_setup, language, timezone, user, \
                                   language_apply, timezone_apply, \
                                   console_setup_apply
 import oem_config.tz
+from oem_config.frontend.base import BaseFrontend
+
+from Timezone import TimezoneMap
 
 UIDIR = '/usr/share/oem-config/qt'
 
-WIDGET_STACK_STEPS = {
-    "step_language": 0,
-    "step_timezone": 1,
-    "step_keyboard": 2,
-    "step_user": 3
-}
-
-WIDGET_STACK_MAX_STEPS = 3
-
 class OEMConfUI(QWidget):
 
     def __init__(self, parent):
@@ -72,13 +66,14 @@ class OEMConfUI(QWidget):
         palette.setBrush(self.backgroundRole(),QBrush(pixmap))
         self.setPalette(palette)
 
-class Frontend:
+class Frontend(BaseFrontend):
     def __init__(self):
+        BaseFrontend.__init__(self)
+
         self.previous_excepthook = sys.excepthook
         sys.excepthook = self.excepthook
 
         self.debconf_callbacks = {}
-        self.locale = None
         self.language_questions = ('oem_config', 'language_label',
                                    'language_heading_label',
                                    'timezone_heading_label',
@@ -86,10 +81,8 @@ class Frontend:
                                    'user_heading_label',
                                    'back', 'next')
         self.current_step = None
-        self.dbfilter = None
         # Set default language.
-        dbfilter = language.Language(self, DebconfCommunicator('oem-config',
-                                                               cloexec=True))
+        dbfilter = language.Language(self, self.debconf_communicator())
         dbfilter.cleanup()
         dbfilter.db.shutdown()
         self.allowed_change_step = True
@@ -140,32 +133,11 @@ class Frontend:
         self.translate_widgets()
 
         self.customize_installer()
-        self.current_layout = None
-        self.map_vbox = QVBoxLayout(self.userinterface.map_frame)
-        self.map_vbox.setMargin(0)
+        
         self.tzmap = TimezoneMap(self)
-        self.tzmap.tzmap.show()
-
-    def post_mortem(self, exctype, excvalue, exctb):
-        """Drop into the debugger if possible."""
-
-        # Did the user request this?
-        if 'OEM_CONFIG_DEBUG_PDB' not in os.environ:
-            return
-        # We must not be in interactive mode; if we are, there's no point.
-        if hasattr(sys, 'ps1'):
-            return
-        # stdin and stdout must point to a terminal. (stderr is redirected
-        # in debug mode!)
-        if not sys.stdin.isatty() or not sys.stdout.isatty():
-            return
-        # SyntaxErrors can't meaningfully be debugged.
-        if issubclass(exctype, SyntaxError):
-            return
-
-        import pdb
-        pdb.post_mortem(exctb)
-        sys.exit(1)
+        map_vbox = QVBoxLayout(self.userinterface.map_frame)
+        map_vbox.setMargin(0)
+        map_vbox.addWidget(self.tzmap)
 
     def excepthook(self, exctype, excvalue, exctb):
         """Crash handler."""
@@ -179,8 +151,6 @@ class Frontend:
         self.previous_excepthook(exctype, excvalue, exctb)
 
     def run(self):
-        global WIDGET_STACK_STEPS, WIDGET_STACK_MAX_STEPS
-
         if os.getuid() != 0:
             title = ('This installer must be run with administrative '
                      'privileges, and cannot continue without them.')
@@ -196,7 +166,6 @@ class Frontend:
         self.app.connect(self.userinterface.language_list, SIGNAL("itemSelectionChanged()"), self.on_language_treeview_selection_changed)
         self.app.connect(self.userinterface.keyboard_list_1, SIGNAL("itemSelectionChanged()"), self.on_keyboard_layout_selected)
         self.app.connect(self.userinterface.keyboard_list_2, SIGNAL("itemSelectionChanged()"), self.on_keyboard_variant_selected)
-        self.app.connect(self.userinterface.city_combo, SIGNAL("activated(int)"), self.tzmap.city_combo_changed)
 
         first_step = "step_language"
         self.userinterface.stackedWidget.setCurrentWidget(self.userinterface.step_language)
@@ -230,7 +199,7 @@ class Frontend:
             else:
                 if self.current_step == 'step_language':
                     self.translate_widgets()
-                self.userinterface.stackedWidget.setCurrentIndex(WIDGET_STACK_STEPS[curr]+1)
+                self.userinterface.stackedWidget.setCurrentIndex(self.pages.index(curr) + 1)
                 self.set_current_page()
             self.app.processEvents()
         if self.apply_changes:
@@ -243,9 +212,12 @@ class Frontend:
             dbfilter = console_setup_apply.ConsoleSetupApply(self)
             dbfilter.run_command(auto_process=True)
 
+            return 0
+        else:
+            return 10
+
 
     def customize_installer(self):
-        global WIDGET_STACK_MAX_STEPS
         self.step_icon_size = QSize(32,32)
         self.step_icons = [self.userinterface.step_icon_lang, self.userinterface.step_icon_loc, \
                            self.userinterface.step_icon_key, self.userinterface.step_icon_user]
@@ -259,7 +231,7 @@ class Frontend:
             self.step_icons_path = ["locale.png", "clock.png", "keyboard_layout.png", "userconfig.png"]
         self.step_labels_text = [self.userinterface.language_heading_label.text(),self.userinterface.timezone_heading_label.text(), \
                                 self.userinterface.keyboard_heading_label.text(), self.userinterface.user_heading_label.text()]
-        for icon in range(WIDGET_STACK_MAX_STEPS+1):
+        for icon in range(len(self.pages)):
             self.step_icons[icon].setPixmap(QPixmap(str(self.step_icons_path_prefix+self.step_icons_path[icon])))
 
     # Internationalisation.
@@ -334,12 +306,6 @@ p, li { white-space: pre-wrap; }
         else:
             print "WARNING: unknown widget: " + name
 
-    def get_string(self, name, lang=None):
-        """Get the string name in the given lang or a default."""
-        if lang is None:
-            lang = self.locale
-        return i18n.get_string(name, lang)
-
 
     def on_keyboard_layout_selected(self):
         if isinstance(self.dbfilter, console_setup.ConsoleSetup):
@@ -364,8 +330,8 @@ p, li { white-space: pre-wrap; }
             return ''
 
     def set_language_choices(self, choices, choice_map):
+        BaseFrontend.set_language_choices(self, choices, choice_map)
         self.userinterface.language_list.clear()
-        self.language_choice_map = dict(choice_map)
         self.lang_store = QStringList()
         for choice in choices:
             self.lang_store.append(QString(choice))
@@ -384,12 +350,11 @@ p, li { white-space: pre-wrap; }
     def on_language_treeview_selection_changed(self):
         lang = self.selected_language()
         if lang:
-            global WIDGET_STACK_MAX_STEPS
             # strip encoding; we use UTF-8 internally no matter what
             lang = lang.split('.')[0].lower()
             for widget in (self.userinterface, self.userinterface.language_label, self.userinterface.welcome_label, self.userinterface.back, self.userinterface.next):
                 self.translate_widget(widget, lang)
-            for step in range(WIDGET_STACK_MAX_STEPS + 1):
+            for step in range(len(self.pages)):
                 self.translate_widget(self.step_labels[step], lang)
                 self.step_labels_text[step] = self.step_labels[step].text()
 
@@ -406,8 +371,12 @@ p, li { white-space: pre-wrap; }
             self.key_store_1.append(QString(choice))
         self.userinterface.keyboard_list_1.addItems(self.key_store_1)
 
-    def set_keyboard(self, keyboard):
-        index = self.key_store_1.indexOf(QRegExp("^"+keyboard+"$"))
+        if self.current_layout is not None:
+            self.set_keyboard(self.current_layout)
+
+    def set_keyboard(self, layout):
+        BaseFrontend.set_keyboard(self, layout)
+        index = self.key_store_1.indexOf(QRegExp("^"+layout+"$"))
         if index != -1:
             self.userinterface.keyboard_list_1.setCurrentRow(index)
 
@@ -430,10 +399,10 @@ p, li { white-space: pre-wrap; }
         return unicode(self.userinterface.keyboard_list_2.currentItem().text())
 
     def set_timezone (self, timezone):
-        self.tzmap.set_tz_from_name(timezone)
+        self.tzmap.set_timezone(timezone)
 
     def get_timezone (self):
-        return self.tzmap.get_selected_tz_name()
+        return self.tzmap.get_timezone()
 
     def set_fullname(self, value):
         self.userinterface.name_ledit.setText(value)
@@ -494,23 +463,7 @@ p, li { white-space: pre-wrap; }
         # TODO cjwatson 2006-02-10: handle dbfilter.status
         self.app.disconnect(self.socketNotifierWrite, SIGNAL("activated(int)"), self.watch_debconf_fd_helper_write)
         self.app.disconnect(self.socketNotifierException, SIGNAL("activated(int)"), self.watch_debconf_fd_helper_exception)
-        if dbfilter is None:
-            name = 'None'
-        else:
-            name = dbfilter.__class__.__name__
-        if self.dbfilter is None:
-            currentname = 'None'
-        else:
-            currentname = self.dbfilter.__class__.__name__
-        syslog.syslog(syslog.LOG_DEBUG,
-                      "debconffilter_done: %s (current: %s)" %
-                      (name, currentname))
-        if dbfilter == self.dbfilter:
-            self.dbfilter = None
-            #if isinstance(dbfilter, summary.Summary):
-                ## The Summary component is just there to gather information,
-                # and won't call run_main_loop() for itself.
-                #self.allow_change_step(True)
+        if BaseFrontend.debconffilter_done(self, dbfilter):
             self.app.exit()
 
     def error_dialog (self, title, msg):
@@ -518,6 +471,35 @@ p, li { white-space: pre-wrap; }
         # TODO: cancel button as well if capb backup
         QMessageBox.warning(self.userinterface, title, msg, QMessageBox.Ok)
 
+    def question_dialog (self, title, msg, options, use_templates=True):
+        # I doubt we'll ever need more than three buttons.
+        assert len(options) <= 3, options
+
+        self.allow_change_step(True)
+        buttons = {}
+        messageBox = QMessageBox(QMessageBox.Question, title, msg, QMessageBox.NoButton, self.userinterface)
+        for option in options:
+            if use_templates:
+                text = self.get_string(option)
+            else:
+                text = option
+            if text is None:
+                text = option
+            # Convention for options is to have the affirmative action last; KDE
+            # convention is to have it first.
+            if option == options[-1]:
+                button = messageBox.addButton(text, QMessageBox.AcceptRole)
+            else:
+                button = messageBox.addButton(text, QMessageBox.RejectRole)
+            buttons[button] = option
+
+        response = messageBox.exec_()
+
+        if response < 0:
+            return None
+        else:
+            return buttons[messageBox.clickedButton()]
+
     def run_main_loop (self):
         if not self.apply_changes:
             self.allow_change_step(True)
@@ -533,7 +515,7 @@ p, li { white-space: pre-wrap; }
         curr = str(self.get_current_step())
         self.backup = True
         if self.dbfilter is not None:
-            self.userinterface.stackedWidget.setCurrentIndex(WIDGET_STACK_STEPS[curr]-1)
+            self.userinterface.stackedWidget.setCurrentIndex(self.pages.index(curr) - 1)
             self.allow_change_step(False)
             self.dbfilter.cancel_handler()
             self.set_current_page()
@@ -546,8 +528,8 @@ p, li { white-space: pre-wrap; }
             self.dbfilter.ok_handler()
 
     def set_current_page(self):
-        global WIDGET_STACK_STEPS, WIDGET_STACK_MAX_STEP
         current_name = self.get_current_step()
+        current_page = self.pages.index(str(current_name))
         if current_name == 'step_language':
             self.userinterface.back.hide()
         else:
@@ -560,15 +542,15 @@ p, li { white-space: pre-wrap; }
         for icon in self.step_icons:
             pixmap = QIcon(icon.pixmap()).pixmap(self.step_icon_size, QIcon.Disabled)
             icon.setPixmap(pixmap)
-        for step in range(WIDGET_STACK_MAX_STEPS+1):
+        for step in range(len(self.pages)):
             self.step_labels_text[step].replace("p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">  <span style=\" font-size:13pt; font-weight:800; font-style:italic;\">","<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">  <span style=\" font-size:13pt; color:gray;\">")
             self.step_labels[step].setText(self.step_labels_text[step])
-        current_icon = self.step_icons[WIDGET_STACK_STEPS[str(current_name)]]
-        current_pixmap = QPixmap(str(self.step_icons_path_prefix+self.step_icons_path[WIDGET_STACK_STEPS[str(current_name)]]))
+        current_icon = self.step_icons[current_page]
+        current_pixmap = QPixmap(str(self.step_icons_path_prefix+self.step_icons_path[current_page]))
         current_icon.setPixmap(current_pixmap)
-        current_label_text = self.step_labels_text[WIDGET_STACK_STEPS[str(current_name)]]
+        current_label_text = self.step_labels_text[current_page]
         current_label_text.replace("<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">  <span style=\" font-size:13pt; color:gray;\">","<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">  <span style=\" font-size:13pt; font-weight:800; font-style:italic;\">")
-        self.step_labels[WIDGET_STACK_STEPS[str(current_name)]].setText(current_label_text)
+        self.step_labels[current_page].setText(current_label_text)
 
     def allow_change_step(self, allowed):
         if allowed:
@@ -584,259 +566,3 @@ p, li { white-space: pre-wrap; }
     def get_current_step(self):
         return self.userinterface.stackedWidget.currentWidget().objectName()
 
-class TimezoneMap(object):
-    def __init__(self, frontend):
-        self.frontend = frontend
-        self.tzdb = oem_config.tz.Database()
-        #self.tzmap = ubiquity.emap.EMap()
-        self.tzmap = MapWidget(self.frontend.userinterface.map_frame)
-        self.frontend.map_vbox.addWidget(self.tzmap)
-        self.tzmap.show()
-        self.update_timeout = None
-        self.point_selected = None
-        self.point_hover = None
-        self.location_selected = None
-
-        timezone_city_combo = self.frontend.userinterface.city_combo
-        self.timezone_city_index = {}  #map human readable city name to Europe/London style zone
-        self.city_index = []  # map cities to indexes for the combo box
-
-        prev_continent = ''
-        for location in self.tzdb.locations:
-            #self.tzmap.add_point("", location.longitude, location.latitude,
-            #                     NORMAL_RGBA)
-            zone_bits = location.zone.split('/')
-            if len(zone_bits) == 1:
-                continue
-            continent = zone_bits[0]
-            if continent != prev_continent:
-                timezone_city_combo.addItem('')
-                self.city_index.append('')
-                timezone_city_combo.addItem("--- %s ---" % continent)
-                self.city_index.append("--- %s ---" % continent)
-                prev_continent = continent
-            human_zone = '/'.join(zone_bits[1:]).replace('_', ' ')
-            timezone_city_combo.addItem(human_zone)
-            self.timezone_city_index[human_zone] = location.zone
-            self.city_index.append(human_zone)
-            self.tzmap.cities[human_zone] = [location.latitude, location.longitude]
-
-        self.frontend.app.connect(self.tzmap, SIGNAL("cityChanged"), self.cityChanged)
-        self.mapped()
-
-    def set_city_text(self, name):
-        """ Gets a long name, Europe/London """
-        timezone_city_combo = self.frontend.userinterface.city_combo
-        count = timezone_city_combo.count()
-        found = False
-        i = 0
-        zone_bits = name.split('/')
-        human_zone = '/'.join(zone_bits[1:]).replace('_', ' ')
-        while not found and i < count:
-            if str(timezone_city_combo.itemText(i)) == human_zone:
-                timezone_city_combo.setCurrentIndex(i)
-                found = True
-            i += 1
-
-    def set_zone_text(self, location):
-        offset = location.utc_offset
-        if offset >= datetime.timedelta(0):
-            minuteoffset = int(offset.seconds / 60)
-        else:
-            minuteoffset = int(offset.seconds / 60 - 1440)
-        if location.zone_letters == 'GMT':
-            text = location.zone_letters
-        else:
-            text = "%s (GMT%+d:%02d)" % (location.zone_letters,
-                                         minuteoffset / 60, minuteoffset % 60)
-        self.frontend.userinterface.tz_selected_label.setText(text)
-        translations = gettext.translation('iso_3166',
-                                           languages=[self.frontend.locale],
-                                           fallback=True)
-        self.frontend.userinterface.region_selected_label.setText(translations.ugettext(location.human_country))
-        self.update_current_time()
-
-    def update_current_time(self):
-        if self.location_selected is not None:
-            now = datetime.datetime.now(self.location_selected.info)
-            self.frontend.userinterface.time_current_label.setText(unicode(now.strftime('%X'), "utf-8"))
-
-    def set_tz_from_name(self, name):
-        """ Gets a long name, Europe/London """
-
-        (longitude, latitude) = (0.0, 0.0)
-
-        for location in self.tzdb.locations:
-            if location.zone == name:
-                (longitude, latitude) = (location.longitude, location.latitude)
-                break
-        else:
-            return
-
-        self.location_selected = location
-        self.set_city_text(self.location_selected.zone)
-        self.set_zone_text(self.location_selected)
-        self.frontend.allow_go_forward(True)
-
-        if name == None or name == "":
-            return
-
-    def get_tz_from_name(self, name):
-        if len(name) != 0:
-            return self.timezone_city_index[name]
-        else:
-            return None
-
-    def city_combo_changed(self, index):
-        city = str(self.frontend.userinterface.city_combo.currentText())
-        try:
-            zone = self.timezone_city_index[city]
-        except KeyError:
-            return
-        self.set_tz_from_name(zone)
-
-    def get_selected_tz_name(self):
-        name = str(self.frontend.userinterface.city_combo.currentText())
-        return self.get_tz_from_name(name)
-
-    def timeout(self):
-        self.update_current_time()
-        return True
-
-    def mapped(self):
-        if self.update_timeout is None:
-            self.update_timeout = QTimer()
-            self.frontend.app.connect(self.update_timeout, SIGNAL("timeout()"), self.timeout)
-            self.update_timeout.start(100)
-
-    def cityChanged(self):
-        self.frontend.userinterface.city_combo.setCurrentIndex(self.city_index.index(self.tzmap.city))
-        self.city_combo_changed(self.frontend.userinterface.city_combo.currentIndex())
-        self.frontend.allow_go_forward(True)
-
-class CityIndicator(QLabel):
-    def __init__(self, parent, name="cityindicator"):
-        QLabel.__init__(self, parent)
-        self.setMouseTracking(True)
-        self.setMargin(1)
-        self.setIndent(0)
-        self.setAutoFillBackground(True)
-        self.setLineWidth(1)
-        self.setFrameStyle(QFrame.Box | QFrame.Plain)
-        self.setPalette(QToolTip.palette())
-        self.setText("CityIndicator")
-
-    def mouseMoveEvent(self, mouseEvent):
-        mouseEvent.ignore()
-
-    def setText(self, text):
-        """ implement auto resize """
-        QLabel.setText(self, text)
-        self.adjustSize()
-
-class MapWidget(QWidget):
-    def __init__(self, parent, name="mapwidget"):
-        QWidget.__init__(self, parent)
-        self.setObjectName(name)
-        self.setAutoFillBackground(True)
-        self.imagePath = "/usr/share/oem-config/pixmaps/world_map-960.png"
-        image = QImage(self.imagePath);
-        pixmapUnscaled = QPixmap(self.imagePath);
-        pixmap = pixmapUnscaled.scaled( QSize(self.width(), self.height()) )
-        palette = QPalette()
-        palette.setBrush(self.backgroundRole(), QBrush(pixmap))
-        self.setPalette(palette)
-        self.cities = {}
-        self.cities['Edinburgh'] = [self.coordinate(False, 55, 50, 0), self.coordinate(True, 3, 15, 0)]
-        self.timer = QTimer(self)
-        self.connect(self.timer, SIGNAL("timeout()"), self.updateCityIndicator)
-        self.setMouseTracking(True)
-
-        self.cityIndicator = CityIndicator(self)
-        self.cityIndicator.setText("")
-        self.cityIndicator.hide()
-
-    def paintEvent(self, paintEvent):
-        ##FIXME this is slow, need to buffer the output.
-        painter = QPainter(self)
-        for city in self.cities:
-            self.drawCity(self.cities[city][0], self.cities[city][1], painter)
-
-    def drawCity(self, lat, long, painter):
-        point = self.getPosition(lat, long, self.width(), self.height())
-        painter.setPen(QPen(QColor(250,100,100), 1))
-        painter.drawPoint(point.x(), point.y()-1)
-        painter.drawPoint(point.x()-1, point.y())
-        painter.drawPoint(point.x(), point.y())
-        painter.drawPoint(point.x()+1, point.y())
-        painter.drawPoint(point.x(), point.y()+1)
-        painter.setPen(QPen(QColor(0,0,0), 1))
-        painter.drawPoint(point.x(), point.y()-2)
-        painter.drawPoint(point.x()-1, point.y()-1)
-        painter.drawPoint(point.x()+1, point.y()-1)
-        painter.drawPoint(point.x()-2, point.y())
-        painter.drawPoint(point.x()+2, point.y())
-        painter.drawPoint(point.x()-1, point.y()+1)
-        painter.drawPoint(point.x()+1, point.y()+1)
-        painter.drawPoint(point.x(), point.y()+2)
-
-    def getPosition(self, la, lo, w, h):
-        x = (w * (180.0 + lo) / 360.0)
-        y = (h * (90.0 - la) / 180.0)
-
-        return QPoint(int(x),int(y))
-
-    def coordinate(self, neg, d, m, s):
-        if neg:
-            return - (d + m/60.0 + s/3600.0)
-        else :
-            return d + m/60.0 + s/3600.0
-
-    def getNearestCity(self, w, h, x, y):
-        result = None
-        dist = 1.0e10
-        for city in self.cities:
-            pos = self.getPosition(self.cities[city][0], self.cities[city][1], self.width(), self.height())
-
-            d = (pos.x()-x)*(pos.x()-x) + (pos.y()-y)*(pos.y()-y)
-            if d < dist:
-                dist = d
-                self.where = pos
-                result = city
-        return result
-
-    def mouseMoveEvent(self, mouseEvent):
-        self.x = mouseEvent.pos().x()
-        self.y = mouseEvent.pos().y()
-        if not self.timer.isActive():
-            self.timer.setSingleShot(True)
-            self.timer.start(25)
-
-    def updateCityIndicator(self):
-        city = self.getNearestCity(self.width(), self.height(), self.x, self.y)
-        if city is None:
-            return
-        self.cityIndicator.setText(city)
-        movePoint = self.getPosition(self.cities[city][0], self.cities[city][1], self.width(), self.height())
-        self.cityIndicator.move(movePoint.x(), movePoint.y() - self.cityIndicator.height())
-        self.cityIndicator.show()
-
-    def mouseReleaseEvent(self, mouseEvent):
-        pos = mouseEvent.pos()
-
-        city = self.getNearestCity(self.width(), self.height(), pos.x(), pos.y());
-        if city is None:
-            return
-        elif city == "Edinburgh":
-            self.city = "London"
-        else:
-            self.city = city
-        self.emit(SIGNAL("cityChanged"), ())
-
-    def resizeEvent(self, resizeEvent):
-        image = QImage(self.imagePath);
-        pixmapUnscaled = QPixmap(self.imagePath);
-        pixmap = pixmapUnscaled.scaled( QSize(self.width(), self.height()), Qt.IgnoreAspectRatio, Qt.SmoothTransformation)
-        palette = QPalette()
-        palette.setBrush(self.backgroundRole(), QBrush(pixmap))
-        self.setPalette(palette)
